
import contextlib
import re
import shlex
import textwrap
import tempfile
import typing as t
from pathlib import Path
import docker  # type: ignore
import pytest

PIP_CACHES_VOLUME = 'pip-caches'
PROJECT_DIRECTORY = str(Path(__file__).parent.parent.resolve())
CUSTOM_STEPS_YAML = '''
steps:
- echo rtd-custom-steps says "$@"
- python -m "${1}" --version
- echo $PYTHON
'''
READTHEDOCS_CONFIG_YAML = '''
python:
  install:
    - requirements: rtd/requirements.txt
'''
PYPROJECT_TOML = '''
[tool.readthedocs-custom-steps]
script = """
echo rtd-custom-steps says "$@"
python -m "${1}" --version
echo $PYTHON
"""
'''


def stop_and_remove_container(container):
  try:
    container.reload()
    if container.status == 'running':
      container.stop()
    container.reload()
    if container.status == 'running':
      container.kill()
  finally:
    container.remove()


@pytest.mark.parametrize(
  argnames='python_image',
  argvalues=[
    '3.6-alpine',
    '3.7-alpine',
    '3.8-alpine',
    '3.9-alpine',
    '3.10-alpine',
  ],
)
@pytest.mark.parametrize(
  argnames='config_type',
  argvalues=['CustomFile:$pwd', 'CustomFile:docs', 'Pyproject:$pwd'],
)
@pytest.mark.parametrize(
  argnames='with_rtd_config',
  argvalues=['WithRtdConfig', 'WithoutRtdConfig'],
)
@pytest.mark.parametrize(
  argnames='tool',
  argvalues=['mkdocs', 'sphinx'],
)
def test_mkdocs_hook(
  python_image: str,
  config_type: str,
  with_rtd_config: bool,
  tool: str,
) -> None:
  """
  This unittest uses Docker to test the readthedocs-custom-steps hook, verifying the output from
  the custom steps as well as the reentry into the hook invoking the same command but not running
  into an infinite loop of invoking itself.
  """

  config_type, config_dir = config_type.split(':')

  if config_type == 'CustomFile':
    steps_content = CUSTOM_STEPS_YAML
    steps_filename = '.readthedocs-custom-steps.yml'
  else:
    steps_content = PYPROJECT_TOML
    steps_filename = 'pyproject.toml'

  if tool == 'mkdocs':
    command = 'mkdocs build --clean --site-dir _site/html --config-file mkdocs.yml'
    requirements = ['mkdocs==1.2.3']
    second_line_regex = r'python -m mkdocs, version .*$'
  else:
    command = 'sphinx -T -b html -d _build/doctrees -D language=en . _build/html'
    requirements = ['sphinx']
    second_line_regex = r'(__main__\.py|sphinx) .*$'

  if config_dir == '$pwd':
    config_dir = '.'

  client = docker.from_env()
  with contextlib.ExitStack() as stack:
    tmpfile = stack.enter_context(tempfile.NamedTemporaryFile('w', delete=False))
    stack.callback(Path(tmpfile.name).unlink)
    tmpfile.write(steps_content)
    tmpfile.close()

    volumes = {
      PIP_CACHES_VOLUME: {'bind': '/root/.cache/pip', 'mode': 'rw'},
      PROJECT_DIRECTORY: {'bind': '/opt/project', 'mode': 'ro'},
      tmpfile.name: {'bind': f'/opt/{steps_filename}', 'mode': 'ro'}}

    if with_rtd_config == 'WithRtdConfig':
      tmpfile = stack.enter_context(tempfile.NamedTemporaryFile('w', delete=False))
      stack.callback(Path(tmpfile.name).unlink)
      tmpfile.write(READTHEDOCS_CONFIG_YAML)
      tmpfile.close()
      volumes[tmpfile.name] = {'bind': '/opt/.readthedocs.yml', 'mode': 'ro'}

    container = client.containers.run(
      'python:' + python_image,
      volumes=volumes,
      command=['sh', '-c', textwrap.dedent(f'''
        set -e
        cp -r /opt/project /tmp/rtdcs
        rm -rf /tmp/rtdcs/*.egg-info /tmp/rtdcs/build

        # Install rtd-cs
        READTHEDOCS=True pip install {" ".join(map(shlex.quote, requirements))} /tmp/rtdcs -v 1>&2

        # Setup test directory
        mkdir /tmp/test; cd /tmp/test;
        if [ -f /opt/.readthedocs.yml ]; then cp /opt/.readthedocs.yml .; fi
        mkdir -p {config_dir}
        cp /opt/{steps_filename} {config_dir}

        # Link shim to test if rtd-cs detects it
        mkdir -p /root/.pyenv/shims
        ln -s "$(which python)" /root/.pyenv/shims/python3.7

        # Invoke build command, which should trigger rtd-cs
        python -m {command}
      ''')],
      detach=True,
      stdin_open=False,
    )

    stack.callback(lambda: stop_and_remove_container(container))

    exit_code = container.wait()['StatusCode']
    result = container.logs().decode()
    print(result)
    if exit_code != 0:
      raise RuntimeError(f'container returned exit code {exit_code}')

    lines = result.splitlines()
    lines = [l for l in lines if not l.startswith('[readthedocs-custom-steps')]
    assert lines[-3] == f'rtd-custom-steps says {command}', lines[-3]
    assert re.match(second_line_regex, lines[-2]) is not None, lines[-2]
    assert lines[-1] == '/root/.pyenv/shims/python3.7', lines[-1]




"""
This script is executed when pydoc-markdown-mkdocs-proxy is installed. This should only run
from inside a readthedocs build process, thus it will assert that READTHEDOCS is actually set.
"""

import argparse
import os
import pipes
import stat
import sys

parser = argparse.ArgumentParser()
parser.add_argument('--dry', action='store_true', help='Do not actually make changes on the filesystem.')
parser.add_argument('--rtd', action='store_true', help='Behave as if READTHEDOCS was set.')
args = parser.parse_args()

if os.getenv('SETUPTOOLS_BUILD') == 'True':
  sys.exit()

if not args.rtd and os.getenv('READTHEDOCS') != 'True':
  raise EnvironmentError(
    'READTHEDOCS must be {!r}, got {!r}.'.format('True', os.getenv('READTHEDOCS')))

new_executable = sys.executable + '-original'
if os.path.isfile(new_executable):
  print('appears to be already installed')
  sys.exit(0)

script = '''#!{SHELL}
log() {{
  >&2 echo "[readthedocs-custom-steps shim]:" "$@"
}}

log "received arguments: $@"

dispatch() {{
  log "dispatching custom steps"
  RTD_CUSTOM_ENTRY=true {python} -c "from readthedocs_custom_steps import main; main()" "$@"
}}

set -e
if [ "$1" = "-m" ] && [ "$2" = "mkdocs" ] && [ "$3" = "build" ] && [ -z "$RTD_CUSTOM_ENTRY" ]; then
  shift
  dispatch "$@"
elif [ "$1" = "-m" ] && [ "$2" = "sphinx" ] && [ -z "$RTD_CUSTOM_ENTRY" ]; then
  shift
  dispatch "$@"
else
  log "forwarding to original Python executable"
  {python} "$@"
fi
'''.format(SHELL=os.getenv('SHELL', '/bin/sh'), python=pipes.quote(new_executable))

print('"{}" -> "{}"'.format(sys.executable, new_executable))
if not args.dry:
  os.rename(sys.executable, new_executable)
print('write "{}" with content'.format(sys.executable))
print('  ' + '\n  '.join(script.splitlines()))
if not args.dry:
  with open(sys.executable, 'w') as fp:
    fp.write(script)
  os.chmod(sys.executable, os.stat(sys.executable).st_mode | stat.S_IEXEC)




import logging
import os
import re
import subprocess
import sys
import textwrap
import typing as t
from pathlib import Path

import tomli
import yaml

__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'
__version__ = '0.6.2'

log = logging.getLogger(__name__)
PYENV_SHIMS_DIR = Path(os.path.expanduser('~')) / '.pyenv/shims'
READTHEDOCS_CONFIG = Path('.readthedocs.yml')
READTHEDOCS_CS_CONFIG = Path('.readthedocs-custom-steps.yml')
PYPROJECT_TOML = Path('pyproject.toml')


def get_referenced_requirements_files() -> t.List[str]:
  """
  Reads the Read the Docs configuration file and extracts all filenames referenced in
  `$.python.install[*].requirements` config values.
  """

  if not READTHEDOCS_CONFIG.exists():
    return []

  config = yaml.safe_load(READTHEDOCS_CONFIG.read_text())
  install = config.get('python', {}).get('install', [])
  requirements: t.List[t.Optional[str]]= [x.get('requirements') for x in install]
  return [x for x in requirements if x]


def find_config_file() -> Path:
  """
  Finds the RtdCS config file.
  """

  choices: t.List[Path] = []

  for directory in [Path('.'), Path('docs')] + [Path(f).parent for f in get_referenced_requirements_files()]:
    path = (Path(directory) / READTHEDOCS_CS_CONFIG).resolve()
    if path.exists():
      return path
    if path not in choices:
      choices.append(path)

  raise RuntimeError(f'file {READTHEDOCS_CS_CONFIG} could not be found, searched in\n- ' + '\n- '.join(map(str, choices)))


def find_pyenv_shims() -> t.Dict[t.Tuple[int, int], str]:
  """
  Finds Python shims in `/home/docs/.pyenv/shims` and returns a dictionary mapping Python major.minor version
  pairs to the full path.
  """

  if not PYENV_SHIMS_DIR.is_dir():
    return {}

  result = {}
  for path in PYENV_SHIMS_DIR.iterdir():
    match = re.match(r'^python3\.(\d+)$', path.name)
    if match:
      result[(3, int(match.group(1)))] = str(path)

  return result


def main():
  argv = sys.argv[1:]
  if argv and argv[0] in ('-v', '--version'):
    print('readthedocs-custom-steps', __version__)
    return

  if PYPROJECT_TOML.exists():
    config = tomli.loads(PYPROJECT_TOML.read_text()).get('tool', {}).get('readthedocs-custom-steps')
    filename = PYPROJECT_TOML
  else:
    config = None
    filename = None

  if config is None:
    filename = find_config_file()
    config = yaml.safe_load(filename.read_text())

  env = os.environ.copy()
  shims = find_pyenv_shims()
  if shims:
    env.update({f'PYTHON{x}{y}': p for (x, y), p in shims.items()})
    env['PYTHON'] = shims[max(shims)]

  shell = os.getenv('SHELL', '/bin/sh')
  bash_script = 'set -e\n'
  if 'steps' in config:
    assert isinstance(config['steps'], list)
    bash_script += '\n'.join(config['steps'])
  elif 'script' in config:
    bash_script += textwrap.dedent(config['script'])
  else:
    raise RuntimeError(f'configuration "{filename}" contains no "script" or "steps" key')

  command = [shell, '-c', bash_script] + sys.argv
  print('[readthedocs-custom-steps dispatch]: running', command, file=sys.stderr)

  sys.exit(subprocess.call(command, env=env))


if __name__ == '__main__':
  main()



# This file was auto-generated by Shut. DO NOT EDIT
# For more information about Shut, check out https://pypi.org/project/shut/

from __future__ import print_function
from setuptools.command.install import install as _install_command
import io
import os
import setuptools
import sys

install_hooks = {
  "before-install": [
    [
      "bin/install-hack.py"
    ]
  ],
  "after-install": [],
  "before-develop": [],
  "after-develop": []
}

def _run_hooks(event):
  import subprocess, shlex, os
  def _shebang(fn):
    with open(fn) as fp:
      line = fp.readline()
      if line.startswith('#'):
        return shlex.split(line[1:].strip())
      return []
  for command in install_hooks[event]:
    if command[0].endswith('.py') or 'python' in _shebang(command[0]):
      command.insert(0, sys.executable)
    env = os.environ.copy()
    env['SHUT_INSTALL_HOOK_EVENT'] = event
    res = subprocess.call(command, env=env)
    if res != 0:
      raise RuntimeError('command {!r} returned exit code {}'.format(command, res))

class install_command(_install_command):
  def run(self):
    _run_hooks('before-install')
    super(install_command, self).run()
    _run_hooks('after-install')

readme_file = 'readme.md'
if os.path.isfile(readme_file):
  with io.open(readme_file, encoding='utf8') as fp:
    long_description = fp.read()
else:
  print("warning: file \"{}\" does not exist.".format(readme_file), file=sys.stderr)
  long_description = None

requirements = [
  'PyYAML >=5.0.0',
  'tomli >=1.2.3',
]
test_requirements = [
  'docker >=5.0.0,<6.0.0',
  'six',
  'types-PyYAML',
]
extras_require = {}
extras_require['test'] = test_requirements

setuptools.setup(
  name = 'readthedocs-custom-steps',
  version = '0.6.2',
  author = 'Niklas Rosenstein',
  author_email = 'rosensteinniklas@gmail.com',
  description = 'A hack to run custom steps when building documentation on Read the Docs.',
  long_description = long_description,
  long_description_content_type = 'text/markdown',
  url = 'https://github.com/NiklasRosenstein/readthedocs-custom-steps/',
  license = 'MIT',
  py_modules = ['readthedocs_custom_steps'],
  package_dir = {'': 'src'},
  include_package_data = True,
  install_requires = requirements,
  extras_require = extras_require,
  tests_require = test_requirements,
  python_requires = '>=3.5.0,<4.0.0',
  data_files = [],
  entry_points = {},
  cmdclass = {'install': install_command},
  keywords = [],
  classifiers = [],
  zip_safe = True,
)
